#!/bin/bash

#########################################################################
# PSXC'S RESCANNER - a wrapper for p-zs's rescan function.              #
##########################################################              #
# This small script will extend project-zs's rescan function. It allows #
# whole sections to be rescanned, or the entire site if needed.         #
#########################################################################

## CONFIG ##

# Enter sitedir - usually no need to change
SITEDIR="/site"

# Enter allowed list of sections (case sensitive). Space separated.
SECTIONS="MP3 0DAY ISOS GAMES XBOX PS2 XXX VCD SVCD DIVX TVSERIES"

# Affils may need to rescan stuff. Enter the affill section(s) here.
# Space separated list. If this variable is empty, affils will have no
# right to rescan.
SEC_AFFILL="AFFILS GROUPS PRE"

# Enter keyword to scan (all) dirs recursively. It will start in current dir
# and rescan every dir below...
SEC_ALL="ALL"

# Flags follow glftpd's standard, except that ! do not work (yet).
# Enter flags for those allowed to rescan current dir. Space separated list.
FLAG_GENERAL="*"

# Enter flags for those allowed to rescan sections. Space separated list.
FLAG_SECTIONS="1 2"

# Enter flags for those allowed to rescan whole site. Space separated list.
FLAG_ALL="1 -psxc =admins"

# Some needed files
PASSWDFILE=/etc/passwd
GROUPFILE=/etc/group
USERDIR=/ftp-data/users

## END OF CONFIG ##

version=0.2
section=$1

check_perms () {
 flaglist="`echo "$@" | sed "s/*/STAR/g"`"
 for flag in $flaglist; do
  if [ "$flag" = "STAR" ]; then
   allowed=1
  elif [ ! -z "`echo "$flag" | grep -e "^!"`" ]; then
   # not currently supported
   flag=""
  elif [ ! -z "`echo "$flag" | grep -e "^-"`" ]; then
   flag="`echo $flag | sed "s/-//"`"
   if [ "$myuname" = "$flag" ]; then
    allowed=1
   fi
  elif [ ! -z "`echo "$flag" | grep -e "^="`" ]; then
   flag="`echo $flag | sed "s/=//"`"
   if [ "$mygname" = "$flag" ]; then
    allowed=1
   fi
  elif [ ! -z "`echo "$myflags" | grep -ie "$flag"`" ]; then
   allowed=1
  fi
 done
 echo -n "$allowed"
}

myuid=`myid | cut -d ' ' -f 1`
mygid=`myid | cut -d ' ' -f 2`
myuname="`cat $PASSWDFILE | cut -d ':' -f 1,3 | tr ':' ' ' | while read uname uid; do if [ $uid -eq $myuid ]; then echo $uname; break; fi; done`"
mygname="`cat $GROUPFILE | cut -d ':' -f 1,3 | tr ':' ' ' | while read gname gid; do if [ $gid -eq $mygid ]; then echo $gname; break; fi; done`"
myflags="`cat $USERDIR/$myuname | grep -e "^FLAGS" | sed "s/FLAGS //" | tr -d ' '`"

allowed_all_perms=0
allowed_sections_perms=0
allowed_current_perms=0
allowed=0

if [ ! -z "$section" ]; then
 if [ ! -z "`echo "$SECTIONS" | grep -e "$section"`" ]; then
  section="$SITEDIR/$section/*"
 elif [ "$section" = "$SEC_ALL" ]; then
  section="$PWD/*"
 fi
else
 section="$PWD"
fi

allowed_all_perms=`check_perms "$FLAG_ALL"`
allowed_all="`echo $allowed_all_perms | sed "s/1/Yes/" | sed "s/0/No/"`"
allowed_sections_perms=`check_perms "$FLAG_SECTIONS"`
allowed_sections="`echo $allowed_sections_perms | sed "s/1/Yes/" | sed "s/0/No/"`"
allowed_current_perms=`check_perms "$FLAG_GENERAL"`
allowed_current="`echo $allowed_current_perms | sed "s/1/Yes/" | sed "s/0/No/"`"

if [ $allowed_current_perms ] && [ ! allowed_all_perms ]; then
 allowed=0
 for affillname in `echo "$SEC_AFFILL" | tr ' ' '\n'`; do 
  if [ ! -z "`echo $PWD | grep -e "$affillname"`" ]; then
   if [ ! -z  "`echo $PWD | grep -e "$mygname"`" ]; then
    allowed=1
    break
   fi
  fi
 done
 if [ $allowed ]; then
  /bin/rescan 2>/dev/null
  exit 0
 fi
fi

if [ $allowed ]; then
 find $section -type d | grep -v "\[" | while read dirname; do
  cd $dirname
  mydirname=`echo $dirname | sed "s|$SITEDIR||"`
  echo -e "\nEntering dir: $mydirname"
  /bin/rescan 2>/dev/null
 done
 echo "Rescan done."
else
 echo "psxc's rescanner v$version"
 echo "---------------------"
 echo "Syntax: site rescan <section>"
 echo ""
 echo "Allowed sections:"
 echo "$SECTIONS"
 echo ""
 echo "You have the right to:"
 echo "Scan current dir: ${allowed_current}."
 echo "Scan sections   : ${allowed_sections}."
 echo "Scan whole site : ${allowed_all}."
fi
exit 0

